# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: weaviate.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0eweaviate.proto\x12\x0cweaviategrpc\x1a\x1cgoogle/protobuf/struct.proto\"\xfb\x02\n\rSearchRequest\x12\x12\n\nclass_name\x18\x01 \x01(\t\x12\r\n\x05limit\x18\x02 \x01(\r\x12\x41\n\x15\x61\x64\x64itional_properties\x18\x03 \x01(\x0b\x32\".weaviategrpc.AdditionalProperties\x12\x33\n\x0bnear_vector\x18\x04 \x01(\x0b\x32\x1e.weaviategrpc.NearVectorParams\x12\x33\n\x0bnear_object\x18\x05 \x01(\x0b\x32\x1e.weaviategrpc.NearObjectParams\x12,\n\nproperties\x18\x06 \x01(\x0b\x32\x18.weaviategrpc.Properties\x12\x37\n\rhybrid_search\x18\x07 \x01(\x0b\x32 .weaviategrpc.HybridSearchParams\x12\x33\n\x0b\x62m25_search\x18\x08 \x01(\x0b\x32\x1e.weaviategrpc.BM25SearchParams\"\xb4\x01\n\x14\x41\x64\x64itionalProperties\x12\x0c\n\x04uuid\x18\x01 \x01(\x08\x12\x0e\n\x06vector\x18\x02 \x01(\x08\x12\x18\n\x10\x63reationTimeUnix\x18\x03 \x01(\x08\x12\x1a\n\x12lastUpdateTimeUnix\x18\x04 \x01(\x08\x12\x10\n\x08\x64istance\x18\x05 \x01(\x08\x12\x11\n\tcertainty\x18\x06 \x01(\x08\x12\r\n\x05score\x18\x07 \x01(\x08\x12\x14\n\x0c\x65xplainScore\x18\x08 \x01(\x08\"]\n\nProperties\x12\x1a\n\x12non_ref_properties\x18\x01 \x03(\t\x12\x33\n\x0eref_properties\x18\x02 \x03(\x0b\x32\x1b.weaviategrpc.RefProperties\"V\n\x12HybridSearchParams\x12\r\n\x05query\x18\x01 \x01(\t\x12\x12\n\nproperties\x18\x02 \x03(\t\x12\x0e\n\x06vector\x18\x03 \x03(\x02\x12\r\n\x05\x61lpha\x18\x04 \x01(\x02\"5\n\x10\x42M25SearchParams\x12\r\n\x05query\x18\x01 \x01(\t\x12\x12\n\nproperties\x18\x02 \x03(\t\"v\n\rRefProperties\x12\x14\n\x0clinked_class\x18\x01 \x01(\t\x12\x1a\n\x12reference_property\x18\x02 \x01(\t\x12\x33\n\x11linked_properties\x18\x03 \x01(\x0b\x32\x18.weaviategrpc.Properties\"l\n\x10NearVectorParams\x12\x0e\n\x06vector\x18\x01 \x03(\x02\x12\x16\n\tcertainty\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x64istance\x18\x03 \x01(\x01H\x01\x88\x01\x01\x42\x0c\n\n_certaintyB\x0b\n\t_distance\"h\n\x10NearObjectParams\x12\n\n\x02id\x18\x01 \x01(\t\x12\x16\n\tcertainty\x18\x02 \x01(\x01H\x00\x88\x01\x01\x12\x15\n\x08\x64istance\x18\x03 \x01(\x01H\x01\x88\x01\x01\x42\x0c\n\n_certaintyB\x0b\n\t_distance\"H\n\x0bSearchReply\x12+\n\x07results\x18\x01 \x03(\x0b\x32\x1a.weaviategrpc.SearchResult\x12\x0c\n\x04took\x18\x02 \x01(\x02\"\x86\x01\n\x0cSearchResult\x12\x32\n\nproperties\x18\x01 \x01(\x0b\x32\x1e.weaviategrpc.ResultProperties\x12\x42\n\x15\x61\x64\x64itional_properties\x18\x02 \x01(\x0b\x32#.weaviategrpc.ResultAdditionalProps\"\xef\x02\n\x15ResultAdditionalProps\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0e\n\x06vector\x18\x02 \x03(\x02\x12\x1a\n\x12\x63reation_time_unix\x18\x03 \x01(\x04\x12\"\n\x1a\x63reation_time_unix_present\x18\x04 \x01(\x08\x12\x1d\n\x15last_update_time_unix\x18\x05 \x01(\x04\x12%\n\x1dlast_update_time_unix_present\x18\x06 \x01(\x08\x12\x10\n\x08\x64istance\x18\x07 \x01(\x02\x12\x18\n\x10\x64istance_present\x18\x08 \x01(\x08\x12\x11\n\tcertainty\x18\t \x01(\x02\x12\x19\n\x11\x63\x65rtainty_present\x18\n \x01(\x08\x12\r\n\x05score\x18\x0b \x01(\x02\x12\x15\n\rscore_present\x18\x0c \x01(\x08\x12\x15\n\rexplain_score\x18\r \x01(\t\x12\x1d\n\x15\x65xplain_score_present\x18\x0e \x01(\x08\"\x91\x01\n\x10ResultProperties\x12\x33\n\x12non_ref_properties\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x34\n\tref_props\x18\x02 \x03(\x0b\x32!.weaviategrpc.ReturnRefProperties\x12\x12\n\nclass_name\x18\x03 \x01(\t\"\\\n\x13ReturnRefProperties\x12\x32\n\nproperties\x18\x01 \x03(\x0b\x32\x1e.weaviategrpc.ResultProperties\x12\x11\n\tprop_name\x18\x02 \x01(\t2N\n\x08Weaviate\x12\x42\n\x06Search\x12\x1b.weaviategrpc.SearchRequest\x1a\x19.weaviategrpc.SearchReply\"\x00\x42#Z!github.com/weaviate/weaviate/grpcb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'weaviate_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z!github.com/weaviate/weaviate/grpc'
  _SEARCHREQUEST._serialized_start=63
  _SEARCHREQUEST._serialized_end=442
  _ADDITIONALPROPERTIES._serialized_start=445
  _ADDITIONALPROPERTIES._serialized_end=625
  _PROPERTIES._serialized_start=627
  _PROPERTIES._serialized_end=720
  _HYBRIDSEARCHPARAMS._serialized_start=722
  _HYBRIDSEARCHPARAMS._serialized_end=808
  _BM25SEARCHPARAMS._serialized_start=810
  _BM25SEARCHPARAMS._serialized_end=863
  _REFPROPERTIES._serialized_start=865
  _REFPROPERTIES._serialized_end=983
  _NEARVECTORPARAMS._serialized_start=985
  _NEARVECTORPARAMS._serialized_end=1093
  _NEAROBJECTPARAMS._serialized_start=1095
  _NEAROBJECTPARAMS._serialized_end=1199
  _SEARCHREPLY._serialized_start=1201
  _SEARCHREPLY._serialized_end=1273
  _SEARCHRESULT._serialized_start=1276
  _SEARCHRESULT._serialized_end=1410
  _RESULTADDITIONALPROPS._serialized_start=1413
  _RESULTADDITIONALPROPS._serialized_end=1780
  _RESULTPROPERTIES._serialized_start=1783
  _RESULTPROPERTIES._serialized_end=1928
  _RETURNREFPROPERTIES._serialized_start=1930
  _RETURNREFPROPERTIES._serialized_end=2022
  _WEAVIATE._serialized_start=2024
  _WEAVIATE._serialized_end=2102
# @@protoc_insertion_point(module_scope)
